from .general import code_or_proprietary, amount_field, address, party, account, agent

def payment_group_header(tag):
    return {
        '_self': tag,
        '_sorting': ['MsgId', 'CreDtTm', 'Authstn', 'NbOfTxs', 'CtrlSum', 'InitgPty', 'FwdgAgt'],
        'message_id': 'MsgId',
        'creation_date_time': 'CreDtTm',
        'authorisation': code_or_proprietary('Authstn'),
        'transaction_count': 'NbOfTxs',
        'control_sum': 'CtrlSum',
        'initiating_party': party('InitgPty'),
        'forwarding_agent': agent('FwdgAgt')
    }

def type_information(tag):
    return {
        '_self': tag,
        '_sorting': ['InstrPrty', 'ClrChanl', 'SvcLvl', 'LclInstrm', 'SeqTp', 'CtgyPurp'],
        'priority': 'InstrPrty',
        'clearing_channel': 'ClrChanl',
        'service_level': code_or_proprietary('SvcLvl'),
        'local_instrument': code_or_proprietary('LclInstrm'),
        'sequence_type': 'SeqTp',
        'category_purpose': code_or_proprietary('CtgyPurp')
    }

def transaction(tag):
    return {
        '_self': tag,
        '_sorting': [
            'PmtId', 'PmtTpInf', 'InstdAmt', 'ChrgBr', 'DrctDbtTx', 'UltmtCdtr', 'DbtrAgt', 'DbtrAgtAcct', 'Dbtr', 'DbtrAcct', 'UltmtDbtr', 'InstrForCdtrAgt',
            'Purp', 'RgltryRptg', 'Tax', 'RltdRmtInf', 'RmtInf'
        ],
        'id': {
            '_self': 'PmtId',
            '_sorting': ['InstrId', 'EndToEndId'],
            'instruction': 'InstrId',
            'end_to_end': 'EndToEndId'
        },
        'type': type_information('PmtTpInf'),
        'amount': amount_field('InstdAmt'),
        'charge_bearer': 'ChrgsBr',
        'transaction': {
            '_self': 'DrctDbtTx',
            '_sorting': ['MndtRltdInf', 'CdtrSchmeId', 'PreNtfctnId', 'PreNtfctnDt'],
            'mandate': {
                '_self': 'MndtRltdInf',
                '_sorting': ['MndtId', 'DtOfSgntr', 'AmdmntInd', 'AmdmntInfDtls', 'ElctmcSgntr', 'FrstColltnDt', 'FnlColltnDt', 'Frqcy', 'Rsn', 'TrckgDays'],
                'id': 'MndtId',
                'date': 'DtOfSgntr',
                'amendment_indicator': 'AmdmntInd',
                'amendment': {
                    '_self': 'AmdmntInfDtls',
                    '_sorting': None
                    # TODO: amendment
                },
                'signature': 'ElctmcSgntr',
                'first_collection': 'FrstColltnDt',
                'final_collection': 'FnlColltnDt',
                'frequency': {
                    '_self': 'Frqcy',
                    '_sorting': ['Tp', 'Prd', 'PtInTm'],
                    'type': 'Tp',
                    'period': {
                        '_self': 'Prd',
                        '_sorting': ['Tp', 'CntPerPrd'],
                        'type': 'Tp',
                        'count': 'CntPerPrd'
                    },
                    'point': {
                        '_self': 'PtInTm',
                        '_sorting': ['Tp', 'PtInTm'],
                        'type': 'Tp',
                        'point': 'PtInTm'
                    }
                },
                'reason': code_or_proprietary('Rsn'),
                'tracking_days': 'TrckgDays'
            },
            'creditor_scheme_identification': party('CdtrSchmeId'),
            'pre_notification_id': 'PreNtfctnId',
            'pre_notification_date': 'PreNtfctnDt'
        },
        'ultimate_creditor': party('UltmtCdtr'),
        'debtor': party('Dbtr'),
        'debtor_account': account('DbtrAcct'),
        'debtor_agent': agent('DbtrAgt'),
        'debtor_agent_account': account('DbtrAgtAcct'),
        'ultimate_debtor': party('UltmtDbtr'),
        'instruction_for_creditor_agent': 'InstrForCdtrAgt',
        'purpose': code_or_proprietary('Purp'),
        'regulatory_reporting': [{
            '_self': 'RgltryRptg',
            '_sorting': ['DbtCdtRptgInd', 'Authrty', 'Dtls'],
            'indicator': 'DbtCdtRptgInd',
            'authority': {
                '_self': 'Authrty',
                '_sorting': ['Nm', 'Ctry'],
                'name': 'Nm',
                'country': 'Ctry'
            },
            'details': {
                '_self': 'Dtls',
                '_sorting': ['Tp', 'Dt', 'Ctry', 'Cd', 'Amt', 'Inf'],
                'type': 'Tp',
                'date': 'Dt',
                'country': 'Ctry',
                'code': 'Cd',
                'amount': 'Amt',
                'information': ['Inf']
            }
        }],
        'tax': {
            '_self': 'Tax',
            '_sorting': None
            # TODO: tax object
        },
        'related_remittance_information': [{
            '_self': 'RltdRmtInf',
            '_sorting': ['RmtId', 'RmtLctnDtls', 'ElctrncAdr', 'PstlAdr'],
            'id': 'RmtId',
            'location': [{
                '_self': 'RmtLctnDtls',
                '_sorting': ['Mtd', 'ElctrncAdr', 'PstlAdr'],
                'method': 'Mtd',
                'electronic_address': 'ElctrncAdr',
                'postal_address': {
                    '_self': 'PstlAdr',
                    '_sorting': ['Nm', 'Adr'],
                    'name': 'Nm',
                    'address': address('Adr')
                }
            }]
        }],
        'remittance_information': {
            '_self': 'RmtInf',
            '_sorting': ['Ustrd', 'Strd'],
            'unstructured': 'Ustrd',
            'structured': {
                '_self': 'Strd',
                '_sorting': ['AddtlRmtInf'],
                # TODO: structured remittance information object,
                'additional_information': ['AddtlRmtInf']
            }
        }
    }

def payment(tag):
    return {
        '_self': tag,
        '_sorting': [
            'PmtInfId', 'PmtMtd', 'BtchBookg', 'NbOfTxs', 'CtrlSum', 'PmtTpInf', 'ReqdColltnDt', 'Cdtr', 'CdtrAcct', 'CdtrAgt', 'CdtrAgtAcct', 'UltmtCdtr',
            'ChrgBr', 'ChrgsAcct', 'ChrgsAcctAgt', 'CdtrSchmeId', 'DrctDbtTxInf', 'SplmtryData'
        ],
        'id': 'PmtInfId',
        'method': 'PmtMtd',
        'batch_booking': 'BtchBookg',
        'transaction_count': 'NbOfTxs',
        'control_sum': 'CtrlSum',
        'type': type_information('PmtTpInf'),
        'collection_date': 'ReqdColltnDt',
        'creditor': party('Cdtr'),
        'creditor_account': account('CdtrAcct'),
        'creditor_agent': agent('CdtrAgt'),
        'creditor_agent_account': account('CdtrAgtAcct'),
        'ultimate_creditor': party('UltmtCdtr'),
        'charge_bearer': 'ChrgBr',
        'charges_account': account('ChrgsAcct'),
        'charges_account_agent': agent('ChrgsAcctAgt'),
        'creditor_scheme_identification': party('CdtrSchmeId'),
        'transactions': [transaction('DrctDbtTxInf')],
        'supplementary_data': ['SplmtryData']
    }



import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import unittest
from lxml import etree
from sepa import builder, parser, validator, signer
from .util import xml_compare

data_object = {
    'group_header': {
        'message_id': '1234567890',
        'creation_date_time': '2017-03-05T13:45:00',
        'authorisation': {
            'code': 'ILEV'
        }
    },
    'mandate': [{
        'id': '78904536',
        'request_id': '9823701',
        'authentication': [{
            'date': '2017-03-05',
            'channel': {
                'code': 'ABC'
            }
        }]
    }]
}

data_xml = ('<MndtInitnReq>'
    '<Mndt>'
        '<Authntcn>'
            '<Dt>2017-03-05</Dt>'
            '<Chanl>'
                '<Cd>ABC</Cd>'
            '</Chanl>'
        '</Authntcn>'
        '<MndtId>78904536</MndtId>'
        '<MndtReqId>9823701</MndtReqId>'
    '</Mndt>'
    '<GrpHdr>'
        '<CreDtTm>2017-03-05T13:45:00</CreDtTm>'
        '<Authstn>'
            '<Cd>ILEV</Cd>'
        '</Authstn>'
        '<MsgId>1234567890</MsgId>'
    '</GrpHdr>'
'</MndtInitnReq>')

class TestSuite(unittest.TestSuite):
    def test_builder(self):
        # Build XML
        data_out = builder.build(builder.mandate_initiation_request, data_object, document=False)

        # Compare generated data to expected output
        assert xml_compare(data_out, etree.fromstring(data_xml))

    def test_validator(self):
        # Build XML
        data_out = builder.build(builder.mandate_initiation_request, data_object)

        # TODO: remove debug print
        print(etree.tostring(data_out))

        # TODO: uncomment this
        # Validate using XML Schema Definition (XSD)
        assert not validator.validate_or_error(validator.mandate_initiation_request, data_out)

    def test_signer(self):
        # Build XML
        data_out = builder.build(builder.mandate_initiation_request, data_object)

        # Sign XML
        with open(os.path.join(os.path.dirname(__file__), 'privkey.pem'), 'rb') as f:
            key = f.read()
        with open(os.path.join(os.path.dirname(__file__), 'cert.pem'), 'rb') as f:
            cert = f.read()
        data_signed = signer.sign(data_out, key=key, cert=cert)

        # TODO: remove debug print
        print(etree.tostring(data_signed))

        # TODO: uncomment this
        # Verify XML
        # assert signer.verify(data_signed)

    def test_parser(self):
        # Parse XML
        data_out = parser.parse(parser.mandate_initiation_request, etree.fromstring(data_xml))

        # Compare generated data to expected output
        assert data_out == data_object

if __name__ == '__main__':
    unittest.main()



import pkgutil
import inspect

sepa_messages = {}

# Loop over modules in this directory
for loader, name, is_pkg in pkgutil.walk_packages(__path__):
    # Load a module
    module = loader.find_module(name).load_module(name)

    # Export the module
    sepa_messages[name] = module



from lxml import etree
import xmltodict

# Stolen from http://stackoverflow.com/a/23368371/4749736

def normalise_dict(d):
    """
    Recursively convert dict-like object (eg OrderedDict) into plain dict.
    Sorts list values.
    """
    out = {}
    for k, v in dict(d).items():
        if hasattr(v, 'items'):
            out[k] = normalise_dict(v)
        elif isinstance(v, list):
            out[k] = []
            for item in sorted(v):
                if hasattr(item, 'items'):
                    out[k].append(normalise_dict(item))
                else:
                    out[k].append(item)
        else:
            out[k] = v
    return out

def xml_compare(a, b):
    """
    Compares two XML documents (as string or etree)

    Does not care about element order
    """
    if not isinstance(a, str):
        a = etree.tostring(a)
    if not isinstance(b, str):
        b = etree.tostring(b)
    a = normalise_dict(xmltodict.parse(a))
    b = normalise_dict(xmltodict.parse(b))
    return a == b



import logging

from lxml import etree
from .messages import sepa_messages

def reverse(structure, name = ''):
    new_structure = {
        '_self': name
    }

    for child in structure:
        if not child.startswith('_'):
            sub = structure[child]

            if isinstance(sub, list):
                if isinstance(sub[0], dict):
                    new_structure[sub[0]['_self']] = [reverse(sub[0], child)]
                else:
                    new_structure[sub[0]] = [child]
            elif isinstance(sub, dict):
                new_structure[sub['_self']] = reverse(sub, child)
            else:
                new_structure[sub] = child
        elif child != '_self':
            new_structure[child] = structure[child]

    return new_structure

# Export parser message types
for group_name, group in sepa_messages.items():
    for name, message in group.items():
        reverse_definition = reverse(message.definition)
        globals()[name] = reverse_definition
        globals()[message.name] = reverse_definition

def parse_tree(structure, tag):
    data = {}

    for child in tag:
        if child.tag not in structure:
            logging.debug('Unknown tag: "' + child.tag + '", parent: "' + tag.tag + '"')
        else:
            substructure = structure[child.tag]

            if isinstance(substructure, list):
                if isinstance(substructure[0], dict):
                    key = substructure[0]['_self']
                    value = parse_tree(substructure[0], child)
                else:
                    key = substructure[0]
                    value = child.text

                if key not in data:
                    data[key] = []
                data[key].append(value)
            elif isinstance(substructure, dict):
                data[substructure['_self']] = parse(substructure, child)

                if '_nochildren' in substructure and substructure['_nochildren']:
                     data[substructure['_self']]['_value'] = child.text

                     if '_attribs' in substructure:
                         for attrib_key, attrib_value in substructure['_attribs'].items():
                             if attrib_value in child.attrib:
                                 data[substructure['_self']][attrib_key] = child.attrib[attrib_value]
            else:
                data[substructure] = child.text

    return data

def parse(structure, tree):
    if tree.tag == 'Document':
        return parse_tree(structure, tree[0])
    return parse_tree(structure, tree)

def parse_string(structure, tree, **kwargs):
    return parse(structure, etree.fromstring(tree, **kwargs))



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()

with open('LICENSE.md', 'r') as f:
    license = f.read()

setup(
    name='sepa',
    version='0.5.1',
    description='Python library for parsing and building SEPA Direct Debit and SEPA eMandate schemas.',
    #long_description=readme,
    #long_description_content_type='text/markdown',
    license=license,
    author='Vereniging Campus Kabel',
    author_email='info@vck.utwente.nl',
    url='https://github.com/VerenigingCampusKabel/python-sepa',
    packages=find_packages(exclude=('tests')),
    # install_requires=['lxml >= 3.5.0, < 4', 'signxml'],
    install_requires=['lxml'],
    test_suite='nose.collector',
    tests_require=['deep', 'xmltodict', 'nose'],
    package_data={
        'sepa': ['schemas/*.xsd']
    }
)
